import abc
import asyncio
import logging
from collections import defaultdict
from decimal import Decimal
from enum import IntEnum
from functools import cached_property
from typing import Awaitable, Dict, List, Optional, Tuple, TypeVar

import a_sync
import brownie
from brownie import ZERO_ADDRESS, chain
from brownie.convert.datatypes import EthAddress
from brownie.exceptions import ContractNotFound
from brownie.network.event import _EventItem

from y import convert
from y.classes.common import ERC20, WeiBalance, _EventsLoader, _Loader
from y.contracts import Contract, contract_creation_block_async
from y.datatypes import (Address, AddressOrContract, AnyAddressType, Block,
                         Pool, UsdPrice, UsdValue)
from y.exceptions import (ContractNotVerified, MessedUpBrownieContract,
                          PriceError, UnsupportedNetwork, call_reverted)
from y.interfaces.curve.CurveRegistry import CURVE_REGISTRY_ABI
from y.networks import Network
from y.utils.events import ProcessedEvents
from y.utils.middleware import ensure_middleware
from y.utils.multicall import \
    multicall_same_func_same_contract_different_inputs
from y.utils.raw_calls import raw_call

T = TypeVar('T')

logger = logging.getLogger(__name__)

ensure_middleware()

# curve registry documentation https://curve.readthedocs.io/registry-address-provider.html
ADDRESS_PROVIDER = '0x0000000022D53366457F9d5E68Ec105046FC4383'

class Ids(IntEnum):
    Main_Registry = 0
    PoolInfo_Getters = 1
    Exchanges = 2
    Metapool_Factory = 3
    Fee_Distributor = 4
    CryptoSwap_Registry = 5
    CryptoPool_Factory = 6
    # On Mainnet, id 7 is listed as "metafactory".
    # On Polygon, id 7 is listed as "cryptopool factory".
    # On other chains, "cryptopool factory" is id 6.
    # On Polygon, id 6 is "crypto factory".
    # I've only seen this on Mainnet and Polygon so far, for now will treat `7` == `6`.
    Cryptopool_Factory = 7
    crvUSD_Plain_Pools_deprecated_1 = 8
    crvUSD_Plain_Pools_deprecated_2 = 9
    crvUSD_Plain_Pools = 10
    Curve_Tricrypto_Factory = 11



class _CurveLoader(_Loader):
    __slots__ = "curve",
    def __init__(self, address: Address, curve: "CurveRegistry", asynchronous: bool = False):
        super().__init__(address, asynchronous=asynchronous)
        self.curve = curve

_LT = TypeVar("_LT", bound=_CurveLoader)

class _CurveEventsLoader(_CurveLoader, _EventsLoader):
    _events: "CurveEvents"

class CurveEvents(ProcessedEvents[_EventItem]):
    __slots__ = "_base",
    def __init__(self, base: _LT):
        super().__init__(addresses=base.address)
        self._base = base
    
class AddressProviderEvents(CurveEvents):
    @property
    def provider(self) -> "AddressProvider":
        return self._base
    def _process_event(self, event) -> None:
        if event.name == 'NewAddressIdentifier' and event['addr'] != ZERO_ADDRESS:
            self.provider.identifiers[Ids(event['id'])].append(event['addr'])
        elif event.name == 'AddressModified' and event['new_address'] != ZERO_ADDRESS:
            self.provider.identifiers[Ids(event['id'])].append(event['new_address'])
        logger.debug("%s loaded event %s at block %s", self, event, event.block_number)
        return event
        
class RegistryEvents(CurveEvents):
    __slots__ = "_tasks",
    def __init__(self, base: _LT):
        super().__init__(base)
        self._tasks: List["asyncio.Task[EthAddress]"] = []
    @property
    def registry(self) -> "Registry":
        return self._base
    def _process_event(self, event: _EventItem) -> None:
        if event.name == 'PoolAdded':
            # TODO async this
            pool = event['pool']
            self._tasks.append(asyncio.create_task(coro=self._add_pool(pool), name=f"Registry._add_pool for pool {pool}"))
            self.registry.curve.registries[event.address].add(pool)
        elif event.name == 'PoolRemoved':
            self.registry.curve.registries[event.address].discard(event['pool'])
        logger.debug("%s loaded event %s at block %s", self, event, event.block_number)
        return event
    async def _add_pool(self, pool: Address) -> EthAddress:
        lp_token = await self.registry.contract.get_lp_token.coroutine(pool)
        self.registry.curve.token_to_pool[lp_token] = pool
    async def _set_lock(self, block: int) -> None:
        await asyncio.gather(*self._tasks)
        self._tasks.clear()
        self._lock.set(block)



class AddressProvider(_CurveEventsLoader):
    __slots__ = "identifiers", "_events", 
    def __init__(self, address: Address, curve: "CurveRegistry", asynchronous: bool = False):
        super().__init__(address, curve, asynchronous=asynchronous)
        self.identifiers = defaultdict(list)
        self._events = AddressProviderEvents(self)
    def __await__(self):
        return self.loaded.__await__()
    async def get_registry(self) -> EthAddress:
        contract = await Contract.coroutine(self.address)
        return await contract.get_registry.coroutine()
    async def _load_factories(self) -> None:
        # factory events are quite useless, so we use a different method
        logger.debug("loading pools from metapool factories")
        metapool_factories = [
            Factory(factory, self.curve, asynchronous=self.asynchronous)
            for i in [Ids.Metapool_Factory, Ids.crvUSD_Plain_Pools, Ids.Curve_Tricrypto_Factory]
            for factory in self.identifiers[i]
        ]

        metapool_factory_pools = await asyncio.gather(*[factory.read_pools(sync=False) for factory in metapool_factories])
        for factory, pool_list in zip(metapool_factories, metapool_factory_pools):
            for pool in pool_list:
                # for metpool factories pool is the same as lp token
                self.curve.token_to_pool[pool] = pool
                self.curve.factories[factory].add(pool)

        # if there are factories that haven't yet been added to the on-chain address provider,
        # please refer to commit 3f70c4246615017d87602e03272b3ed18d594d3c to see how to add them manually
        logger.debug("loading pools from cryptopool factories")
        await asyncio.gather(*[
            Factory(factory, self.curve, asynchronous=self.asynchronous) 
            for factory in self.identifiers[Ids.CryptoPool_Factory] + self.identifiers[Ids.Cryptopool_Factory]
        ])
        if not self.curve._done.is_set():
            logger.info('loaded %s pools from %s registries and %s factories', len(self.curve.token_to_pool), len(self.curve.registries), len(self.curve.factories))
            self.curve._done.set()

class Registry(_CurveEventsLoader):
    __slots__ = "_events"
    def __init__(self, address: Address, curve: "CurveRegistry", asynchronous: bool = False):
        super().__init__(address, curve, asynchronous=asynchronous)
        self._events = RegistryEvents(self)
    def __await__(self):
        return self.loaded.__await__()

class Factory(_CurveLoader):
    def __await__(self):
        return self.loaded.__await__()
    @property
    def loaded(self) -> Awaitable[T]:
        self._task  # ensure task is running and not errd
        if self._loaded is None:
            self._loaded = a_sync.Event(name=f"curve factory {self.address}")
        return self._loaded.wait()
    async def get_pool(self, i: int) -> EthAddress:
        contract = await Contract.coroutine(self.address)
        return await contract.pool_list.coroutine(i)
    async def pool_count(self, block: Optional[int] = None) -> int:
        contract = await Contract.coroutine(self.address)
        return await contract.pool_count.coroutine(block_identifier=block)
    async def read_pools(self) -> List[EthAddress]:
        try:
            # lets load the contract async and then we can use the sync property more conveniently
            await Contract.coroutine(self.address)
        except ContractNotVerified:
            if chain.id == Network.xDai:
                Contract.from_abi("Vyper_contract", self.address, CURVE_REGISTRY_ABI)
            else:
                # This happens sometimes, not sure why as the contract is verified.
                brownie.Contract.from_explorer(self.address)
        return await asyncio.gather(*[self.get_pool(i) for i in range(await self.pool_count())])
    async def _load(self) -> None:
        pool_list = await self.read_pools(sync=False)
        await asyncio.gather(*[self._load_pool(pool) for pool in pool_list if pool not in self.curve.factories[self.address]])
        logger.debug("loaded %s pools for %s", len(pool_list), self)
        self._loaded.set()
    async def _load_pool(self, pool: Address) -> None:
        factory = await Contract.coroutine(self.address)
        # for curve v5 pools, pool and lp token are separate
        if hasattr(factory, 'get_token'):
            lp_token = await factory.get_token.coroutine(pool)
        elif hasattr(factory, 'get_lp_token'):
            lp_token = await factory.get_lp_token.coroutine(pool)
        else:
            raise NotImplementedError(f"New factory {factory.address} is not yet supported. Please notify a ypricemagic maintainer.")
        self.curve.token_to_pool[lp_token] = pool
        self.curve.factories[factory.address].add(pool)
        logger.debug("loaded %s pool %s", self, pool)
        


class CurvePool(ERC20): # this shouldn't be ERC20 but works for inheritance for now
    @a_sync.aka.cached_property
    async def factory(self) -> Contract:
        return await curve.get_factory(self, sync=False)

    @a_sync.aka.cached_property
    async def coins(self) -> List[ERC20]:
        """
        Get coins of pool.
        """
        factory = await self.__factory__(sync=False)
        if factory:
            coins = await factory.get_coins.coroutine(self.address)
        else:
            registry = await curve.__registry__(sync=False)
            coins = await registry.get_coins.coroutine(self.address)
        
        # pool not in registry
        if set(coins) == {ZERO_ADDRESS}:
            coins = await multicall_same_func_same_contract_different_inputs(
                self.address, 
                'coins(uint256)(address)', 
                inputs = [i for i in range(8)],
                return_None_on_failure=True,
                sync=False,
            )

        return [ERC20(coin, asynchronous=self.asynchronous) for coin in coins if coin not in {None, ZERO_ADDRESS}]
    
    @a_sync.a_sync(ram_cache_maxsize=256)
    async def get_coin_index(self, coin: AnyAddressType) -> int:
        return [i for i, _coin in enumerate(await self.__coins__(sync=False)) if _coin == coin][0]
    
    @a_sync.aka.cached_property
    async def num_coins(self) -> int:
        return len(await self.__coins__(sync=False))
    
    async def get_dy(self, coin_ix_in: int, coin_ix_out: int, block: Optional[Block] = None, ignore_pools: Tuple[Pool, ...] = ()) -> Optional[WeiBalance]:
        tokens = await self.__coins__(sync=False)
        token_in = tokens[coin_ix_in]
        token_out = tokens[coin_ix_out]
        amount_in = await token_in.__scale__(sync=False)
        contract = await Contract.coroutine(self.address)
        try:
            amount_out = await contract.get_dy.coroutine(coin_ix_in, coin_ix_out, amount_in, block_identifier=block)
            return WeiBalance(amount_out, token_out, block=block, ignore_pools=(*ignore_pools, self))
        except Exception as e:
            if call_reverted(e):
                return None
            raise
    
    @a_sync.aka.cached_property
    async def coins_decimals(self) -> List[int]:
        factory = await self.__factory__(sync=False)
        source = factory or await curve.registry
        coins_decimals = await source.get_decimals.coroutine(self.address)

        # pool not in registry
        if not any(coins_decimals):
            coins = await self.__coins__(sync=False)
            coins_decimals = await asyncio.gather(*[coin.__decimals__(sync=False) for coin in coins])
        
        return [dec for dec in coins_decimals if dec != 0]
    
    @a_sync.aka.cached_property
    async def get_underlying_coins(self) -> List[ERC20]:    
        factory = await self.__factory__(sync=False)    
        if factory:
            # new factory reverts for non-meta pools
            if not hasattr(factory, 'is_meta') or factory.is_meta(self.address):
                coins = await factory.get_underlying_coins.coroutine(self.address)
            else:
                coins = await factory.get_coins.coroutine(self.address)
        else:
            registry = await curve.registry
            coins = await registry.get_underlying_coins.coroutine(self.address)
        
        # pool not in registry, not checking for underlying_coins here
        if set(coins) == {ZERO_ADDRESS}:
            return await self.__coins__(sync=False)

        return [ERC20(coin, asynchronous=self.asynchronous) for coin in coins if coin != ZERO_ADDRESS]
    
    @a_sync.a_sync(ram_cache_maxsize=1000)
    async def get_balances(self, block: Optional[Block] = None) -> Dict[ERC20, Decimal]:
        """
        Get {token: balance} of liquidity in the pool.
        """

        # TODO figure out why these can't be gathered.
        # Sometimes `self.coins` is a list not a coroutine?
        #coins, decimals = await gather([
        #    self.__coins__(sync=False),
        #    self.__coins_decimals__(sync=False),
        #])

        coins = await self.__coins__(sync=False)
        decimals = await self.__coins_decimals__(sync=False)

        try:
            factory = await self.__factory__(sync=False)
            source = factory or await curve.__registry__(sync=False)
            balances = await source.get_balances.coroutine(self.address, block_identifier=block)
        # fallback for historical queries where registry was not yet deployed
        except ValueError:
            balances = await asyncio.gather(*[self._get_balance(i, block) for i, _ in enumerate(coins)])

        if not any(balances):
            raise ValueError(f'could not fetch balances {self.__str__()} at {block}')

        return {
            coin: Decimal(balance / 10 ** dec)
            for coin, balance, dec in zip(coins, balances, decimals)
            if coin != ZERO_ADDRESS
        }
    
    async def _get_balance(self, i: int, block: Optional[Block] = None) -> Optional[int]:
        contract = await Contract.coroutine(self.address)
        try:
            return await contract.balances.coroutine(i, block_identifier=block)
        except ValueError as e:
            if str(e) == "No data was returned - the call likely reverted":
                return None
            raise

    async def get_tvl(self, block: Optional[Block] = None) -> Optional[UsdValue]:
        """
        Get total value in Curve pool.
        """
        try:
            balances = await self.get_balances(block=block, sync=False)
        except ValueError as e:
            if str(e).startswith("could not fetch balances "):
                return None
            raise e
        
        prices = await asyncio.gather(*[coin.price(block=block, sync=False) for coin in balances])

        return UsdValue(
            sum(balance * Decimal(price) for balance, price in zip(balances.values(), prices))
        )
    
    @a_sync.a_sync(ram_cache_maxsize=100_000, ram_cache_ttl=60*60)
    async def check_liquidity(self, token: Address, block: Block) -> int:
        deploy_block = await contract_creation_block_async(self.address)
        if block < deploy_block:
            return 0
        i = await self.get_coin_index(token, sync=False)
        return await self._get_balance(i, block)
    
    #@cached_property
    #def oracle(self) -> Optional[Address]:
    #    '''
    #    If `pool` has method `price_oracle`, returns price_oracle address.
    #    Else, returns `None`.
    #    '''
    #    response = raw_call(self.address, 'price_oracle()', output='address', return_None_on_failure=True)
    #    if response == ZERO_ADDRESS:
    #        return None
    #    return response



class CurveRegistry(a_sync.ASyncGenericSingleton):
    __slots__ = "__task", 
    def __init__(self, asynchronous: bool = False) -> None:
        super().__init__()
        self.asynchronous = asynchronous
        try: 
            self.address_provider = AddressProvider(ADDRESS_PROVIDER, self, asynchronous=self.asynchronous)
            self.address_provider.contract
        except (ContractNotFound, ContractNotVerified) as e:
            raise UnsupportedNetwork("curve is not supported on this network") from e
        except MessedUpBrownieContract as e:
            if chain.id == Network.Cronos:
                raise UnsupportedNetwork("curve is not supported on this network") from e
            else:
                raise

        self.registries = defaultdict(set)  # registry -> pools
        self.factories = defaultdict(set)  # factory -> pools
        self.pools = set()
        self.token_to_pool = dict()  # lp_token -> pool
    
    def __repr__(self) -> str:
        return "<CurveRegistry>"

    @property
    def identifiers(self) -> List[EthAddress]:
        return self.address_provider.identifiers
    @a_sync.aka.cached_property
    async def registry(self) -> brownie.Contract:
        try:
            return await Contract.coroutine(self.identifiers[0][-1])
        except IndexError: # if we couldn't get the registry via logs
            return await Contract.coroutine(await raw_call(self.address_provider, 'get_registry()', output='address', sync=False))
    
    async def load_all(self) -> None:
        await self._done.wait()

    async def get_factory(self, pool: AddressOrContract) -> Contract:
        """
        Get metapool factory that has spawned a pool.
        """
        try:
            factory = next(
                factory
                for factory, factory_pools in self.factories.items()
                if str(pool) in factory_pools
            )
            return await Contract.coroutine(factory)
        except StopIteration:
            return None    
    
    @a_sync.a_sync(cache_type='memory')
    async def get_price(self, token: Address, block: Optional[Block] = None) -> Optional[float]:
        pool = await self.get_pool(token, sync=False)
        tvl = await pool.get_tvl(block=block, sync=False)
        if tvl is None:
            return None
        return tvl / await ERC20(token, asynchronous=True).total_supply_readable(block)

    @a_sync.a_sync(cache_type='memory')
    async def get_pool(self, token: AnyAddressType) -> CurvePool:
        """
        Get Curve pool (swap) address by LP token address. Supports factory pools.
        """
        await self.load_all()
        
        token = convert.to_address(token)
        if token in self.token_to_pool and token != ZERO_ADDRESS:
            return CurvePool(self.token_to_pool[token], asynchronous=self.asynchronous)

    @a_sync.a_sync(cache_type='memory')
    async def get_price_for_underlying(self, token_in: Address, block: Optional[Block] = None, ignore_pools: Tuple[Pool, ...] = ()) -> Optional[UsdPrice]:
        try:
            pools: List[CurvePool] = (await self.__coin_to_pools__(sync=False))[token_in]
        except KeyError:
            return None
        
        pools = [pool for pool in pools if pool not in ignore_pools]

        if block is not None:
            deploy_blocks = await asyncio.gather(*[contract_creation_block_async(pool.address, True) for pool in pools])
            pools = [pool for pool, deploy_block in zip(pools, deploy_blocks) if deploy_block <= block]

        # Choose a pool to use for pricing `token_in`.
        if len(pools) == 1:
            pool = pools[0]
        else:
            # Use the pool with deepest liquidity.
            balances = await asyncio.gather(*[pool.check_liquidity(token_in, block=block, sync=False) for pool in pools], return_exceptions=True)
            deepest_pool, deepest_bal = None, 0
            for pool, depth in zip(pools, balances):
                if depth > deepest_bal:
                    deepest_pool = pool
                    deepest_bal = depth
            pool = deepest_pool

        if pool is None:
            return None

        if len(coins := await pool.__coins__(sync=False)) != 2:
            # TODO: handle this sitch if necessary
            return

        # Get the price for `token_in` using the selected pool.
        # this works for most typical metapools

        token_in_ix = await pool.get_coin_index(token_in, sync=False)
        token_out_ix = 0 if token_in_ix == 1 else 1 if token_in_ix == 0 else None
        dy = await pool.get_dy(token_in_ix, token_out_ix, block=block, ignore_pools=ignore_pools, sync=False)
        if dy is None:
            return None

        try:
            return await dy.__value_usd__(sync=False)
        except PriceError as e:
            logger.debug("%s for %s at block %s", e.__class__.__name__, token_in, block)
            return None
    
    @a_sync.aka.cached_property
    async def coin_to_pools(self) -> Dict[str, List[CurvePool]]:
        mapping = defaultdict(set)
        await self.load_all()
        pools = {CurvePool(pool) for pools in self.factories.values() for pool in pools}
        for pool in pools:
            for coin in await pool.__coins__(sync=False):
                mapping[coin].add(pool)
        return {coin: list(pools) for coin, pools in mapping.items()}
    
    async def check_liquidity(self, token: Address, block: Block, ignore_pools: Tuple[Pool, ...]) -> int:
        pools: List[CurvePool] = await self.__coin_to_pools__(sync=False)
        if token not in pools:
            return 0
        if pools := [pool for pool in pools[token] if pool not in ignore_pools]:
            return max(await asyncio.gather(*[pool.check_liquidity(token, block, sync=False) for pool in pools]))
        return 0
    
    @cached_property
    def _done(self) -> a_sync.Event:
        """A helper function to ensure the Event is attached to the correct loop."""
        self._task
        return a_sync.Event(name="curve")
    
    @cached_property
    def _task(self) -> asyncio.Task:
        logger.debug("creating loader task for %s", self)
        task = asyncio.create_task(coro=self._load_all(), name=f"{self}._load_all()")
        def done_callback(t: asyncio.Task):
            if e := t.exception():
                logger.error("exception while loading %s: %s", self, e)
                logger.exception(e)
                self.__task = None
                raise e
        task.add_done_callback(done_callback)
        return task

    async def _load_all(self) -> None:
        await self.address_provider
        logger.debug("curve address provider events loaded, now loading factories and pools")
        # NOTE: Gnosis chain's address provider fails to provide registry via events. Maybe other chains as well.
        if not self.identifiers[Ids.Main_Registry]:
            self.identifiers[Ids.Main_Registry] = [await self.address_provider.get_registry()]
        while True:
            # Check if any registries were updated, then ensure all old and new are loaded
            await asyncio.gather(*[
                Registry(self.identifiers[i][-1], self, asynchronous=self.asynchronous)
                for i in [Ids.Main_Registry, Ids.CryptoSwap_Registry]
                if self.identifiers[i]
            ])
            # load metapool and curve v5 factories
            await self.address_provider._load_factories()
            await asyncio.sleep(600)
   
try: 
    curve = CurveRegistry(asynchronous=True)
except UnsupportedNetwork: 
    curve = set()




from tests.fixtures import async_test, mainnet_only
from y import get_price
from y.datatypes import UsdPrice


@async_test
@mainnet_only
async def test_balancer_v2_pool_price():
    v2_weighted_pool = "0x96646936b91d6B9D7D0c47C496AfBF3D6ec7B6f8" # 50/50 WETH/USDC
    assert await get_price(v2_weighted_pool, 14_000_000, sync=False) == UsdPrice(62.35369906899849)

@async_test
@mainnet_only
async def test_balancer_v2_token_price():
    v2_token = "0x616e8BfA43F920657B3497DBf40D6b1A02D4608d"
    assert await get_price(v2_token, 17_500_000, sync=False) == UsdPrice(12.522495269157522)



import logging
from datetime import datetime
from typing import Optional

from a_sync import a_sync
from brownie import chain
from pony.orm import db_session

from y._db.common import token_attr_threads
from y._db.entities import Block, Chain, insert, retry_locked
from y._db.utils._ep import _get_get_block

logger = logging.getLogger(__name__)

@a_sync(default='async', executor=token_attr_threads)
@db_session
@retry_locked
def get_chain() -> Chain:
    return Chain.get(id=chain.id) or insert(type=Chain, id=chain.id) or Chain.get(id=chain.id)

@a_sync(default='async', executor=token_attr_threads)
@db_session
@retry_locked
def get_block(number: int) -> Block:
    chain = get_chain(sync=True)
    if block := Block.get(chain=chain, number=number):
        return block
    return insert(type=Block, chain=chain, number=number) or get_block(number, sync=True)

@a_sync(default='async', executor=token_attr_threads)
@db_session
@retry_locked
def get_block_timestamp(number: int) -> Optional[int]:
    get_block = _get_get_block()
    if ts := get_block(number, sync=True).timestamp:
        return ts.timestamp()
    
@a_sync(default='async', executor=token_attr_threads)
@db_session
@retry_locked
def set_block_timestamp(block: int, timestamp: int) -> None:
    get_block = _get_get_block()
    get_block(block, sync=True).timestamp = datetime.fromtimestamp(timestamp)



import asyncio
import contextlib
import logging
from typing import Any, Callable, Iterable, List, Optional, Tuple, Union

import a_sync
import brownie
from brownie import chain
from eth_abi.exceptions import InsufficientDataBytes
from multicall import Call
from web3.exceptions import CannotHandleRequest

from y import convert
from y.contracts import Contract, contract_creation_block
from y.datatypes import Address, AddressOrContract, AnyAddressType, Block
from y.decorators import stuck_coro_debugger
from y.exceptions import continue_if_call_reverted
from y.interfaces.multicall2 import MULTICALL2_ABI
from y.networks import Network
from y.utils.dank_mids import dank_w3
from y.utils.raw_calls import _decimals, _totalSupply

logger = logging.getLogger(__name__)

MULTICALL2 = {
    Network.Mainnet:            "0x5BA1e12693Dc8F9c48aAD8770482f4739bEeD696",
    Network.Arbitrum:           "0x842eC2c7D803033Edf55E478F461FC547Bc54EB2",
    Network.Avalanche:          "0xdf2122931FEb939FB8Cf4e67Ea752D1125e18858",
    Network.BinanceSmartChain:  "0xfF6FD90A470Aaa0c1B8A54681746b07AcdFedc9B",
    Network.Fantom:             "0xBAD2B082e2212DE4B065F636CA4e5e0717623d18",
    Network.Harmony:            "0x34b415f4d3b332515e66f70595ace1dcf36254c5",
    Network.Heco:               "0xd1F3BE686D64e1EA33fcF64980b65847aA43D79C",
    Network.Moonriver:          "0xaeF00A0Cf402D9DEdd54092D9cA179Be6F9E5cE3",
    Network.Polygon:            "0xc8E51042792d7405184DfCa245F2d27B94D013b6",
    Network.xDai:               "0x9903f30c1469d8A2f415D4E8184C93BD26992573",
    Network.Aurora:             "0xe0e3887b158F7F9c80c835a61ED809389BC08d1b",
    Network.Cronos:             "0x5e954f5972EC6BFc7dECd75779F10d848230345F",
    Network.Optimism:           "0xcA11bde05977b3631167028862bE2a173976CA11", # Multicall 3
    Network.Base:               "0xcA11bde05977b3631167028862bE2a173976CA11", # mc3
}.get(chain.id)

multicall = None
multicall2 = brownie.Contract.from_abi("Multicall2",MULTICALL2, MULTICALL2_ABI) if chain.id in [Network.Harmony,Network.Cronos] else Contract(MULTICALL2)

# the address doesn't matter, it just needs to have code
code = " "

multicall_deploy_block = contract_creation_block(multicall2.address)


@a_sync.a_sync(default='sync')
@stuck_coro_debugger
async def multicall_same_func_no_input(
    addresses: Iterable[AnyAddressType],
    method: str, 
    block: Optional[Block] = None,
    apply_func: Optional[Callable] = None,
    return_None_on_failure: bool = False
    ) -> List[Any]:

    addresses = _clean_addresses(addresses)
    results = await asyncio.gather(*[Call(address, [method], [[address,apply_func]], block_id=block).coroutine() for address in addresses])
    return [v for call in results for k, v in call.items()]


@a_sync.a_sync(default='sync')
@stuck_coro_debugger
async def multicall_same_func_same_contract_different_inputs(
    address: AnyAddressType, 
    method: str, 
    inputs: Union[List, Tuple],  
    block: Optional[Block] = None,
    apply_func: Optional[Callable] = None,
    return_None_on_failure: bool = False
    ) -> List[Any]:
    assert inputs
    address = convert.to_address(address)
    results = await asyncio.gather(
        *[Call(address, [method, input], [[input, apply_func]], block_id=block).coroutine() for input in inputs],
        return_exceptions=return_None_on_failure,
    )
    if return_None_on_failure:
        for i, result in enumerate(results[:]):
            if isinstance(result, Exception):
                continue_if_call_reverted(result)
                results[i] = {i: None}
    return [result for call in results for key, result in call.items()]


@a_sync.a_sync(default='sync')
@stuck_coro_debugger
async def multicall_decimals(
    addresses: Iterable[AddressOrContract], 
    block: Optional[Block] = None,
    return_None_on_failure: bool = True
    ) -> List[int]:

    try: 
        return await asyncio.gather(*[Call(str(address), ['decimals()(uint256)'], block_id=block).coroutine() for address in addresses])
    except (CannotHandleRequest, InsufficientDataBytes):
        pass # TODO investigate these
    except Exception as e:
        continue_if_call_reverted(e)

    return await asyncio.gather(*[_decimals(address,block=block,return_None_on_failure=return_None_on_failure) for address in addresses])

@a_sync.a_sync(default='sync')
@stuck_coro_debugger
async def multicall_totalSupply(
    addresses: Iterable[AddressOrContract], 
    block: Optional[Block] = None,
    return_None_on_failure: bool = True
    ) -> List[int]:

    with contextlib.suppress(CannotHandleRequest, InsufficientDataBytes):
        return await multicall_same_func_no_input(addresses, 'totalSupply()(uint256)', block=block, sync=False)
    return [await _totalSupply(address,block=block,return_None_on_failure=return_None_on_failure, sync=False) for address in addresses] 


#yLazyLogger(logger)
@a_sync.a_sync(default='sync')
@stuck_coro_debugger
async def fetch_multicall(*calls: Any, block: Optional[Block] = None) -> List[Optional[Any]]:
    # https://github.com/makerdao/multicall
    multicall_input = []
    fn_list = []
    decoded = []

    for contract, fn_name, *fn_inputs in calls:
        fn = getattr(contract, fn_name)

        # check that there aren't multiple functions with the same name
        if hasattr(fn, "_get_fn_from_args"):
            fn = fn._get_fn_from_args(fn_inputs)

        fn_list.append(fn)
        multicall_input.append((contract, fn.encode_input(*fn_inputs)))

    if isinstance(block, int) and block < multicall_deploy_block:
        # use state override to resurrect the contract prior to deployment
        data = multicall2.tryAggregate.encode_input(False, multicall_input)
        call = await dank_w3.eth.call(
            {'to': str(multicall2), 'data': data},
            block or 'latest',
            {str(multicall2): {'code': f'0x{multicall2.bytecode}'}},
        )
        result = multicall2.tryAggregate.decode_output(call)
    else:
        result = await multicall2.tryAggregate.coroutine(
            False, multicall_input, block_identifier=block or 'latest'
        )

    for fn, (ok, data) in zip(fn_list, result):
        try:
            assert ok, "call failed"
            decoded.append(fn.decode_output(data))
        except (AssertionError, InsufficientDataBytes):
            decoded.append(None)

    return decoded


#yLazyLogger(logger)
def _clean_addresses(
    addresses: Iterable[AnyAddressType]
    ) -> List[Address]:

    return [convert.to_address(address) for address in addresses]



MULTICALL2_ABI = [{"inputs":[{"components":[{"internalType":"address","name":"target","type":"address"},{"internalType":"bytes","name":"callData","type":"bytes"}],"internalType":"struct Multicall2.Call[]","name":"calls","type":"tuple[]"}],"name":"aggregate","outputs":[{"internalType":"uint256","name":"blockNumber","type":"uint256"},{"internalType":"bytes[]","name":"returnData","type":"bytes[]"}],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"components":[{"internalType":"address","name":"target","type":"address"},{"internalType":"bytes","name":"callData","type":"bytes"}],"internalType":"struct Multicall2.Call[]","name":"calls","type":"tuple[]"}],"name":"blockAndAggregate","outputs":[{"internalType":"uint256","name":"blockNumber","type":"uint256"},{"internalType":"bytes32","name":"blockHash","type":"bytes32"},{"components":[{"internalType":"bool","name":"success","type":"bool"},{"internalType":"bytes","name":"returnData","type":"bytes"}],"internalType":"struct Multicall2.Result[]","name":"returnData","type":"tuple[]"}],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"uint256","name":"blockNumber","type":"uint256"}],"name":"getBlockHash","outputs":[{"internalType":"bytes32","name":"blockHash","type":"bytes32"}],"stateMutability":"view","type":"function"},{"inputs":[],"name":"getBlockNumber","outputs":[{"internalType":"uint256","name":"blockNumber","type":"uint256"}],"stateMutability":"view","type":"function"},{"inputs":[],"name":"getCurrentBlockCoinbase","outputs":[{"internalType":"address","name":"coinbase","type":"address"}],"stateMutability":"view","type":"function"},{"inputs":[],"name":"getCurrentBlockDifficulty","outputs":[{"internalType":"uint256","name":"difficulty","type":"uint256"}],"stateMutability":"view","type":"function"},{"inputs":[],"name":"getCurrentBlockGasLimit","outputs":[{"internalType":"uint256","name":"gaslimit","type":"uint256"}],"stateMutability":"view","type":"function"},{"inputs":[],"name":"getCurrentBlockTimestamp","outputs":[{"internalType":"uint256","name":"timestamp","type":"uint256"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"address","name":"addr","type":"address"}],"name":"getEthBalance","outputs":[{"internalType":"uint256","name":"balance","type":"uint256"}],"stateMutability":"view","type":"function"},{"inputs":[],"name":"getLastBlockHash","outputs":[{"internalType":"bytes32","name":"blockHash","type":"bytes32"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"bool","name":"requireSuccess","type":"bool"},{"components":[{"internalType":"address","name":"target","type":"address"},{"internalType":"bytes","name":"callData","type":"bytes"}],"internalType":"struct Multicall2.Call[]","name":"calls","type":"tuple[]"}],"name":"tryAggregate","outputs":[{"components":[{"internalType":"bool","name":"success","type":"bool"},{"internalType":"bytes","name":"returnData","type":"bytes"}],"internalType":"struct Multicall2.Result[]","name":"returnData","type":"tuple[]"}],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"bool","name":"requireSuccess","type":"bool"},{"components":[{"internalType":"address","name":"target","type":"address"},{"internalType":"bytes","name":"callData","type":"bytes"}],"internalType":"struct Multicall2.Call[]","name":"calls","type":"tuple[]"}],"name":"tryBlockAndAggregate","outputs":[{"internalType":"uint256","name":"blockNumber","type":"uint256"},{"internalType":"bytes32","name":"blockHash","type":"bytes32"},{"components":[{"internalType":"bool","name":"success","type":"bool"},{"internalType":"bytes","name":"returnData","type":"bytes"}],"internalType":"struct Multicall2.Result[]","name":"returnData","type":"tuple[]"}],"stateMutability":"nonpayable","type":"function"}]


import asyncio
import logging
from contextlib import suppress
from decimal import Decimal
from typing import Optional, Tuple

import a_sync
from brownie import chain
from multicall.call import Call

from y import Network
from y.classes.common import ERC20
from y.contracts import Contract, has_method, has_methods, probe
from y.datatypes import AnyAddressType, Block, Pool, UsdPrice
from y.exceptions import (CantFetchParam, ContractNotVerified,
                          MessedUpBrownieContract)
from y.utils.logging import get_price_logger
from y.utils.raw_calls import raw_call

logger = logging.getLogger(__name__)

# NOTE: Yearn and Yearn-inspired

underlying_methods = [
    'token()(address)',
    'underlying()(address)',
    'native()(address)',
    'want()(address)',
    'input()(address)',
    'asset()(address)',
    'wmatic()(address)',
    'wbnb()(address)',
    'based()(address)',
]

share_price_methods = [
    'pricePerShare()(uint)',
    'getPricePerShare()(uint)',
    'getPricePerFullShare()(uint)',
    'getSharesToUnderlying()(uint)',
    'exchangeRate()(uint)',
]

force_false = {
    Network.Mainnet: [
        "0x8751D4196027d4e6DA63716fA7786B5174F04C15",  # wibBTC
        "0xF0a93d4994B3d98Fb5e3A2F90dBc2d69073Cb86b",  # PWRD
    ],
}.get(chain.id, [])

@a_sync.a_sync(default='sync', cache_type='memory')
async def is_yearn_vault(token: AnyAddressType) -> bool:
    # wibbtc returns True here even though it doesn't meet the criteria.
    # TODO figure out a better fix. For now I need a fix asap so this works.
    if chain.id == Network.Mainnet and str(token) == "0x8751D4196027d4e6DA63716fA7786B5174F04C15":
        return False

    # Yearn-like contracts can use these formats
    result = any(
        await asyncio.gather(
            has_methods(token, ('pricePerShare()(uint)','getPricePerShare()(uint)','getPricePerFullShare()(uint)','getSharesToUnderlying()(uint)'), any, sync=False),
            has_methods(token, ('exchangeRate()(uint)','underlying()(address)'), sync=False),
        )
    )

    # pricePerShare can revert if totalSupply == 0, which would cause `has_methods` to return `False`,
    # but it might still be a vault. This section will correct `result` for problematic vaults.
    if not result:
        with suppress(ContractNotVerified, MessedUpBrownieContract):
            contract = await Contract.coroutine(token)
            result = any([
                hasattr(contract, 'pricePerShare'),
                hasattr(contract, 'getPricePerShare'),
                hasattr(contract, 'getPricePerFullShare'),
                hasattr(contract, 'getSharesToUnderlying'),
                hasattr(contract, 'convertToAssets'),
            ])

    return result

@a_sync.a_sync(default='sync')
async def get_price(
    token: AnyAddressType, 
    block: Optional[Block] = None,
    ignore_pools: Tuple[Pool, ...] = (),
) -> UsdPrice:
    return await YearnInspiredVault(token).price(block=block, ignore_pools=ignore_pools, sync=False)

class YearnInspiredVault(ERC20):
    __slots__ = "_get_share_price", 
    # v1 vaults use getPricePerFullShare scaled to 18 decimals
    # v2 vaults use pricePerShare scaled to underlying token decimals
    # yearnish clones use all sorts of other things, we gotchu covered
    def __init__(self, address: AnyAddressType, asynchronous: bool = False):
        super().__init__(address, asynchronous=asynchronous)
        self._get_share_price = None
    
    @a_sync.aka.cached_property
    async def underlying(self) -> ERC20:
        # special cases
        if chain.id == Network.Arbitrum and self.address == '0x57c7E0D43C05bCe429ce030132Ca40F6FA5839d7':
            return ERC20(await raw_call(self.address, 'usdl()', output='address', sync=False), asynchronous=self.asynchronous)

        try:
            underlying = await probe(self.address, underlying_methods)
        except AssertionError:
            # special handler for some strange beefy vaults
            if not (method := {"BeefyVaultV6Matic": "wmatic()", "BeefyVenusVaultBNB": "wbnb()"}.get(self.build_name)):
                raise
            underlying = await raw_call(self.address, method, output='address', sync=False)
        
        if not underlying:
            # certain reaper vaults
            lend_platform = await self.has_method('lendPlatform()(address)', return_response=True, sync=False)
            if lend_platform:
                underlying = await has_method(lend_platform, 'underlying()(address)', return_response=True, sync=False)

        if underlying: 
            return ERC20(underlying, asynchronous=self.asynchronous)
        raise CantFetchParam(f'underlying for {self.__repr__()}')

    a_sync.a_sync(cache_type='memory', ram_cache_maxsize=1000)
    async def share_price(self, block: Optional[Block] = None) -> Optional[Decimal]:
        if self._get_share_price:
            try:
                share_price = await self._get_share_price.coroutine(block_id=block)
            except Exception as e:
                logger.debug("exc %s when fetching share price for %s", e, self)
                share_price = await probe(self.address, share_price_methods, block=block)
        else:
            share_price_method, share_price = await probe(self.address, share_price_methods, block=block, return_method=True)
            if share_price_method:
                self._get_share_price = Call(self.address, [share_price_method])

        if share_price is None:
            # this is for element vaults and other 'scaled' share price functions. probe fails because method requires input
            try:
                contract = await Contract.coroutine(self.address)
                for method in ['convertToAssets', 'getSharesToUnderlying']:
                    if hasattr(contract, method):
                        share_price = await getattr(contract, method).coroutine(await self.__scale__(sync=False), block_identifier=block)
            except ContractNotVerified:
                pass

        if share_price is not None:
            if self._get_share_price and self._get_share_price.function == 'getPricePerFullShare()(uint)':
                # v1 vaults use getPricePerFullShare scaled to 18 decimals
                return share_price / Decimal(10 ** 18)
            underlying: ERC20 = await self.__underlying__(sync=False)
            return Decimal(share_price) / await underlying.__scale__(sync=False)
            
        elif await raw_call(self.address, 'totalSupply()', output='int', block=block, return_None_on_failure=True, sync=False) == 0:
            return None
        
        else:
            raise CantFetchParam(f'share_price for {self.__repr__()}')
    
    a_sync.a_sync(cache_type='memory', ram_cache_maxsize=1000)
    async def price(
        self, 
        block: Optional[Block] = None,
        ignore_pools: Tuple[Pool, ...] = (),
    ) -> UsdPrice:
        logger = get_price_logger(self.address, block=None, extra='yearn')
        underlying: ERC20
        share_price, underlying = await asyncio.gather(self.share_price(block=block, sync=False), self.__underlying__(sync=False))
        if share_price is None:
            return None
        logger.debug("%s share price at block %s: %s", self, block, share_price)
        price = UsdPrice(share_price * Decimal(await underlying.price(block=block, ignore_pools=ignore_pools, sync=False)))
        logger.debug("%s price at block %s: %s", self, block, price)
        return price



import asyncio
import functools
import logging
from typing import (Awaitable, Callable, Iterable, List, NoReturn, Optional,
                    Tuple)

import a_sync
from brownie import ZERO_ADDRESS, chain
from brownie.exceptions import ContractNotFound
from multicall.utils import raise_if_exception_in

from y import ENVIRONMENT_VARIABLES as ENVS
from y import constants, convert
from y.classes.common import ERC20
from y.datatypes import AnyAddressType, Block, Pool, UsdPrice
from y.decorators import stuck_coro_debugger
from y.exceptions import NonStandardERC20, PriceError, yPriceMagicError
from y.prices import convex, one_to_one, popsicle, rkp3r, solidex, yearn
from y.prices.band import band
from y.prices.chainlink import chainlink
from y.prices.dex import mooniswap
from y.prices.dex.balancer import balancer_multiplexer
from y.prices.dex.genericamm import generic_amm
from y.prices.dex.uniswap import uniswap_multiplexer
from y.prices.dex.uniswap.uniswap import uniswap_multiplexer
from y.prices.dex.uniswap.v2 import UniswapV2Pool
from y.prices.eth_derivs import creth, wsteth
from y.prices.gearbox import gearbox
from y.prices.lending import ib
from y.prices.lending.aave import aave
from y.prices.lending.compound import compound
from y.prices.stable_swap import (belt, ellipsis, froyo, mstablefeederpool,
                                  saddle)
from y.prices.stable_swap.curve import curve
from y.prices.synthetix import synthetix
from y.prices.tokenized_fund import basketdao, gelato, piedao, tokensets
from y.prices.utils import ypriceapi
from y.prices.utils.buckets import check_bucket
from y.prices.utils.sense_check import _sense_check
from y.utils.dank_mids import dank_w3
from y.utils.logging import get_price_logger

cache_logger = logging.getLogger(f"{__name__}.cache")

@a_sync.a_sync(default='sync')
async def get_price(
    token_address: AnyAddressType,
    block: Optional[Block] = None,
    *,
    fail_to_None: bool = False,
    skip_cache: bool = ENVS.SKIP_CACHE,
    ignore_pools: Tuple[Pool, ...] = (),
    silent: bool = False,
    ) -> Optional[UsdPrice]:
    '''
    Don't pass an int like `123` into `token_address` please, that's just silly.
    - ypricemagic accepts ints to allow you to pass `y.get_price(0x0bc529c00C6401aEF6D220BE8C6Ea1667F6Ad93e)`
        so you can save yourself some keystrokes while testing in a console
    - (as opposed to `y.get_price("0x0bc529c00C6401aEF6D220BE8C6Ea1667F6Ad93e")`)

    When `get_price` is unable to find a price:
    - If `silent == True`, ypricemagic will print an error message using standard python logging
    - If `silent == False`, ypricemagic will not log any error
    - If `fail_to_None == True`, ypricemagic will return `None`
    - If `fail_to_None == False`, ypricemagic will raise a PriceError
    '''
    block = block or await dank_w3.eth.block_number
    token_address = convert.to_address(token_address)
    try:
        return await _get_price(token_address, block, fail_to_None=fail_to_None, ignore_pools=ignore_pools, skip_cache=skip_cache, silent=silent)
    except (ContractNotFound, NonStandardERC20, PriceError) as e:
        symbol = await ERC20(token_address, asynchronous=True).symbol
        if not fail_to_None:
            raise yPriceMagicError(e, token_address, block, symbol) from e

GetPrice = Callable[..., Awaitable[Optional[UsdPrice]]]



@a_sync.a_sync(default='sync')
async def get_prices(
    token_addresses: Iterable[AnyAddressType],
    block: Optional[Block] = None,
    fail_to_None: bool = False,
    skip_cache: bool = ENVS.SKIP_CACHE,
    silent: bool = False,
    ) -> List[Optional[float]]:
    '''
    A more optimized way to fetch prices for multiple assets at the same block.

    # NOTE silent kwarg is currently disabled.
    In every case:
    - if `silent == True`, tqdm will not be used
    - if `silent == False`, tqdm will be used

    When `get_prices` is unable to find a price:
    - if `fail_to_None == True`, ypricemagic will return `None` for that token
    - if `fail_to_None == False`, ypricemagic will raise a yPriceMagicError
    '''

    if block is None:
        block = chain.height

    prices = await asyncio.gather(
        *[
            get_price(
                token_address=convert.to_address(token), 
                block=block, 
                fail_to_None=fail_to_None, 
                skip_cache=skip_cache, 
                silent=silent, 
                sync=False,
            )
            for token in token_addresses
        ],
        return_exceptions=True
    )

    if not fail_to_None:
        raise_if_exception_in(prices)
    else:
        for p in prices:
            if isinstance(p, Exception) and not isinstance(p, PriceError):
                raise p
    return prices

def __cache(get_price: GetPrice) -> GetPrice:
    @functools.wraps(get_price)
    async def cache_wrap(
        token: AnyAddressType, 
        block: Block,
        *,
        fail_to_None: bool = False, 
        skip_cache: bool = ENVS.SKIP_CACHE,
        ignore_pools: Tuple[Pool, ...] = (),
        silent: bool = False
    ) -> Optional[UsdPrice]:
        from y._db.utils import price as db
        if not skip_cache and (price := await db.get_price(token, block)):
            cache_logger.debug('disk cache -> %s', price)
            return price
        price = await get_price(token, block=block, fail_to_None=fail_to_None, ignore_pools=ignore_pools, silent=silent)
        if price and not skip_cache:
            await db.set_price(token, block, price)
        return price
    return cache_wrap

@stuck_coro_debugger
@a_sync.a_sync(default="async", cache_type='memory', ram_cache_ttl=ENVS.CACHE_TTL)
@__cache
async def _get_price(
    token: AnyAddressType, 
    block: Block,
    *,
    fail_to_None: bool = False, 
    ignore_pools: Tuple[Pool, ...] = (),
    silent: bool = False
    ) -> Optional[UsdPrice]:  # sourcery skip: remove-redundant-if

    try:
        # We do this to cache the symbol for later, otherwise some repr woudl break
        symbol = await ERC20(token, asynchronous=True).symbol
    except NonStandardERC20:
        symbol = None

    logger = get_price_logger(token, block, 'magic')
    logger.debug('fetching price for %s', symbol)
    logger._debugger = asyncio.create_task(coro=_debug_tsk(symbol, logger), name=f"_debug_tsk({symbol}, {logger})")

    # Helps to detect stuck code
    try:
        if token == ZERO_ADDRESS:
            _fail_appropriately(logger, symbol, fail_to_None=fail_to_None, silent=silent)
            logger._debugger.cancel()
            return None

        if ypriceapi.should_use and token not in ypriceapi.skip_tokens:
            price = await ypriceapi.get_price(token, block)
            logger.debug("ypriceapi -> %s", price)
            if price is not None:
                logger.debug("%s price: %s", symbol, price)
                logger._debugger.cancel()
                return price

        price = await _exit_early_for_known_tokens(token, block=block, ignore_pools=ignore_pools, logger=logger)
        if price is not None:
            await _sense_check(token, block, price)
            logger.debug("%s price: %s", symbol, price)
            logger._debugger.cancel()
            return price
        
        # TODO We need better logic to determine whether to use uniswap, curve, balancer. For now this works for all known cases.
        if price is None:
            dexes = [uniswap_multiplexer]
            if curve:
                dexes.append(curve)
            liquidity = await asyncio.gather(*[dex.check_liquidity(token, block, ignore_pools=ignore_pools, sync=False) for dex in dexes])
            depth_to_dex = dict(zip(liquidity, dexes))
            dexes_by_depth = {depth: depth_to_dex[depth] for depth in sorted(depth_to_dex, reverse=True) if depth}
            logger.debug('dexes by depth: %s', dexes_by_depth)
            for dex in dexes_by_depth.values():
                method = 'get_price_for_underlying' if hasattr(dex, 'get_price_for_underlying') else 'get_price'
                logger.debug("trying %s", dex)
                price = await getattr(dex, method)(token, block, ignore_pools=ignore_pools, sync=False)
                logger.debug("%s -> %s", dex, price)
                if price:
                    break

        logger.debug('no %s liquidity found on primary markets', token)

        # If price is 0, we can at least try to see if balancer gives us a price. If not, its probably a shitcoin.
        if price is None or price == 0:
            new_price = await balancer_multiplexer.get_price(token, block=block, sync=False)
            logger.debug("balancer -> %s", price)
            if new_price:
                logger.debug("replacing price %s with new price %s", price, new_price)
                price = new_price

        if price is None:
            _fail_appropriately(logger, symbol, fail_to_None=fail_to_None, silent=silent)
        if price:
            await _sense_check(token, block, price)

        logger.debug("%s price: %s", symbol, price)
        # Don't need this anymore
        logger._debugger.cancel()
        return price
    except Exception as e:
        logger._debugger.cancel()
        raise e


@stuck_coro_debugger
async def _exit_early_for_known_tokens(
    token_address: str,
    block: Block,
    logger: logging.Logger,
    ignore_pools: Tuple[Pool, ...] = (),
    ) -> Optional[UsdPrice]:  # sourcery skip: low-code-quality

    bucket = await check_bucket(token_address, sync=False)

    price = None

    if bucket == 'atoken':                  price = await aave.get_price(token_address, block=block, sync=False)
    elif bucket == 'balancer pool':         price = await balancer_multiplexer.get_price(token_address, block, sync=False)
    elif bucket == 'basketdao':             price = await basketdao.get_price(token_address, block, sync=False)

    elif bucket == 'belt lp':               price = await belt.get_price(token_address, block, sync=False)
    elif bucket == 'chainlink and band':    price = await chainlink.get_price(token_address, block, sync=False) or await band.get_price(token_address, block, sync=False)
    elif bucket == 'chainlink feed':        price = await chainlink.get_price(token_address, block, sync=False)

    elif bucket == 'compound':              price = await compound.get_price(token_address, block=block, sync=False)
    elif bucket == 'convex':                price = await convex.get_price(token_address,block, sync=False)
    elif bucket == 'creth':                 price = await creth.get_price_creth(token_address, block, sync=False)

    elif bucket == 'curve lp':              price = await curve.get_price(token_address, block, sync=False)
    elif bucket == 'ellipsis lp':           price = await ellipsis.get_price(token_address, block=block, sync=False)
    elif bucket == 'froyo':                 price = await froyo.get_price(token_address, block=block, sync=False)

    elif bucket == 'gearbox':               price = await gearbox.get_price(token_address, block=block, sync=False)
    elif bucket == 'gelato':                price = await gelato.get_price(token_address, block=block, sync=False)
    elif bucket == 'generic amm':           price = await generic_amm.get_price(token_address, block=block, sync=False)
    
    elif bucket == 'ib token':              price = await ib.get_price(token_address, block=block, sync=False)
    elif bucket == 'mooniswap lp':          price = await mooniswap.get_pool_price(token_address, block=block, sync=False)
    elif bucket == 'mstable feeder pool':   price = await mstablefeederpool.get_price(token_address,block=block, sync=False)
    
    elif bucket == 'one to one':            price = await one_to_one.get_price(token_address, block, sync=False)
    elif bucket == 'piedao lp':             price = await piedao.get_price(token_address, block=block, sync=False)
    elif bucket == 'popsicle':              price = await popsicle.get_price(token_address, block=block, sync=False)
    
    elif bucket == 'rkp3r':                 price = await rkp3r.get_price(token_address, block, sync=False)
    elif bucket == 'saddle':                price = await saddle.get_price(token_address, block, sync=False)
    elif bucket == 'solidex':               price = await solidex.get_price(token_address, block, sync=False)

    elif bucket == 'stable usd':            price = 1
    elif bucket == 'synthetix':             price = await synthetix.get_price(token_address, block, sync=False)
    elif bucket == 'token set':             price = await tokensets.get_price(token_address, block=block, sync=False)

    elif bucket == 'uni or uni-like lp':    price = await UniswapV2Pool(token_address).get_price(block=block, sync=False)
    elif bucket == 'wrapped gas coin':      price = await get_price(constants.WRAPPED_GAS_COIN, block, sync=False)
    elif bucket == 'wrapped atoken v2':     price = await aave.get_price_wrapped_v2(token_address, block, sync=False)

    elif bucket == 'wrapped atoken v3':     price = await aave.get_price_wrapped_v3(token_address, block, sync=False)
    elif bucket == 'wsteth':                price = await wsteth.wsteth.get_price(block, sync=False)
    elif bucket == 'yearn or yearn-like':   price = await yearn.get_price(token_address, block, ignore_pools=ignore_pools, sync=False)

    logger.debug("%s -> %s", bucket, price)
    return price

         
def _fail_appropriately(
    logger: logging.Logger,
    symbol: str, 
    fail_to_None: bool = False, 
    silent: bool = False
    ) -> None:
    '''
    dictates how `magic.get_price()` will handle failures

    when `get_price` is unable to find a price:
        if `silent == True`, ypricemagic will print an error message using standard python logging
        if `silent == False`, ypricemagic will not log any error
        if `fail_to_None == True`, ypricemagic will return `None`
        if `fail_to_None == False`, ypricemagic will raise a PriceError
    '''
    if not silent:
        logger.warning(f"failed to get price for {symbol}")

    if not fail_to_None:
        raise PriceError(logger, symbol)

async def _debug_tsk(symbol: str, logger: logging.Logger) -> NoReturn:
    """Prints a log every 1 minute until the creating coro returns"""
    while True:
        await asyncio.sleep(60)
        logger.debug("price still fetching for %s", symbol)




import asyncio
from decimal import Decimal
from typing import Optional

import a_sync
from brownie.convert.datatypes import EthAddress
from multicall import Call

from y.classes.common import ERC20, WeiBalance
from y.datatypes import Block, UsdPrice


@a_sync.a_sync(default='sync')
async def is_basketdao_index(address: EthAddress) -> bool:
    try:
        if await Call(address, 'getAssetsAndBalances()(address[],uint[])').coroutine():
            return True
    except:
        return False

@a_sync.a_sync(default='sync')
async def get_price(address: EthAddress, block: Optional[Block] = None) -> UsdPrice:
    balances, total_supply = await asyncio.gather(
        Call(address, 'getAssetsAndBalances()(address[],uint[])',block_id=block).coroutine(),
        ERC20(address, asynchronous=True).total_supply_readable(block=block),
    )

    balances = [
        WeiBalance(balance, token, block)
        for token, balance
        in zip(balances[0],balances[1])
    ]

    tvl = sum(await asyncio.gather(*[bal.__value_usd__(sync=False) for bal in balances]))
    return UsdPrice(tvl / Decimal(total_supply))



import logging
from typing import List, TypeVar, Union

from brownie import chain
from cachetools.func import ttl_cache
from lazy_logging import LazyLoggerFactory
from typing_extensions import ParamSpec

from y.datatypes import AnyAddressType, Block
from y.networks import Network

T = TypeVar('T')
P = ParamSpec('P')

yLazyLogger = LazyLoggerFactory("YPRICEMAGIC")

logger = logging.getLogger(__name__)

@ttl_cache(ttl=10*60)
def get_price_logger(token_address: AnyAddressType, block: Block, extra: str = '') -> logging.Logger:
    address = str(token_address)
    name = f"y.prices.{Network.label()}.{address}.{block}"
    if extra: 
        name += f".{extra}"
    logger = logging.getLogger(name)
    logger.address = address
    logger.block = block
    if logger.level != logger.parent.level:
        logger.setLevel(logger.parent.level)
    return logger

def enable_debug_logging(logger: str = 'y') -> None:
    logger = logging.getLogger(logger)
    logger.setLevel(logging.DEBUG)
    if not logger.handlers:
        logger.addHandler(logging.StreamHandler())

NETWORK_DESCRIPTOR_FOR_ISSUE_REQ =f'name ({Network.name()})' if Network.name() else f'chainid ({chain.id})'

def _gh_issue_request(issue_request_details: Union[str, List[str]], _logger = None) -> None:

    if _logger is None: _logger = logger

    if type(issue_request_details) == str:
        _logger.warning(issue_request_details)

    elif type(issue_request_details) == list:
        for message in issue_request_details:
            _logger.warning(message)

    _logger.warning('Please create an issue and/or create a PR at https://github.com/BobTheBuidler/ypricemagic')
    _logger.warning(f'In your issue, please include the network {NETWORK_DESCRIPTOR_FOR_ISSUE_REQ} and the detail shown above.')
    _logger.warning('and I will add it soon :). This will not prevent ypricemagic from fetching price for this asset.')



import asyncio
import logging
import math
from collections import defaultdict
from functools import cached_property, lru_cache
from itertools import cycle
from typing import AsyncIterator, DefaultDict, Dict, List, Optional, Tuple

import a_sync
from brownie import chain
from brownie.network.event import _EventItem
from eth_abi.packed import encode_abi_packed

from y import ENVIRONMENT_VARIABLES as ENVS
from y.classes.common import ERC20, ContractBase
from y.constants import usdc, weth
from y.contracts import Contract, contract_creation_block_async
from y.datatypes import Address, AnyAddressType, Block, Pool, UsdPrice
from y.decorators import stuck_coro_debugger
from y.exceptions import ContractNotVerified, TokenNotFound, UnsupportedNetwork
from y.interfaces.uniswap.quoterv3 import UNIV3_QUOTER_ABI
from y.networks import Network
from y.utils.events import ProcessedEvents
from y.utils.multicall import fetch_multicall

# https://github.com/Uniswap/uniswap-v3-periphery/blob/main/deploys.md
UNISWAP_V3_FACTORY = '0x1F98431c8aD98523631AE4a59f267346ea31F984'
UNISWAP_V3_QUOTER = '0xb27308f9F90D607463bb33eA1BeBb41C27CE5AB6'

logger = logging.getLogger(__name__)

# same addresses on all networks
addresses = {
    Network.Mainnet: {
        'factory': UNISWAP_V3_FACTORY,
        'quoter': UNISWAP_V3_QUOTER,
        'fee_tiers': [3000, 500, 10_000, 100],
    },
    Network.Arbitrum: {
        'factory': UNISWAP_V3_FACTORY,
        'quoter': UNISWAP_V3_QUOTER,
        'fee_tiers': [3000, 500, 10_000],
    },
    Network.Optimism: {
        'factory': UNISWAP_V3_FACTORY,
        'quoter': UNISWAP_V3_QUOTER,
        'fee_tiers': [3000, 500, 10_000, 100],
    },
    Network.Base: {
        'factory': '0x33128a8fC17869897dcE68Ed026d694621f6FDfD',
        'quoter': '0x3d4e44Eb1374240CE5F1B871ab261CD16335B76a', # quoter v2
        'fee_tiers': [3000, 500, 10_000, 100],
    }
}

FEE_DENOMINATOR = 1_000_000


class UniswapV3Pool(ContractBase):
    __slots__ = 'fee', 'token0', 'token1', 'tick_spacing'
    def __init__(
        self,
        address: Address,
        token0: Address, 
        token1: Address, 
        tick_spacing: int, 
        fee: int, 
        deploy_block: int,
        asynchronous: bool = True
    ) -> None:
        super().__init__(address, asynchronous=asynchronous)
        self.token0 = ERC20(token0, asynchronous=asynchronous)
        self.token1 = ERC20(token1, asynchronous=asynchronous)
        self.tick_spacing = tick_spacing
        self.fee = fee
        self._deploy_block = deploy_block

    def __contains__(self, token: Address) -> bool:
        return token in [self.token0, self.token1]
    
    def __getitem__(self, token: Address) -> ERC20:
        if token not in self:
            raise TokenNotFound(token, self)
        return ERC20(token, asynchronous=self.asynchronous)

    @a_sync.a_sync(ram_cache_maxsize=100_000, ram_cache_ttl=60*60)
    async def check_liquidity(self, token: AnyAddressType, block: Block) -> Optional[int]:
        logger.debug("checking %s liquidity for %s at %s", self, token, block)
        if block < await self.deploy_block(sync=False):
            logger.debug("block %s prior to %s deploy block", block, self)
            return 0
        liquidity = await self[token].balance_of(self.address, block, sync=False)
        logger.debug("%s liquidity for %s at %s: %s", self, token, block, liquidity)
        return liquidity
    
    @lru_cache
    def _get_token_out(self, token_in: ERC20) -> ERC20:
        if token_in == self.token0:
            return self.token1
        elif token_in == self.token1:
            return self.token0
        raise TokenNotFound(token_in, self)


class UniswapV3(a_sync.ASyncGenericSingleton):
    def __init__(self, asynchronous: bool = True) -> None:
        self.asynchronous = asynchronous
        if chain.id not in addresses:
            raise UnsupportedNetwork('compound is not supported on this network')
        self.fee_tiers = addresses[chain.id]['fee_tiers']
        self.loading = False
        self._pools = {}

    def __contains__(self, asset) -> bool:
        return chain.id in addresses

    @a_sync.aka.property
    async def factory(self) -> Contract:
        return await Contract.coroutine(addresses[chain.id]['factory'])
    
    @cached_property
    def loaded(self) -> a_sync.Event:
        return a_sync.Event(name="uniswap v3")
    
    @a_sync.aka.cached_property
    async def quoter(self) -> Contract:
        quoter = addresses[chain.id]['quoter']
        try:
            return await Contract.coroutine(quoter)
        except ContractNotVerified:
            return Contract.from_abi("Quoter", quoter, UNIV3_QUOTER_ABI)
    
    @stuck_coro_debugger
    @a_sync.a_sync(cache_type='memory', ram_cache_ttl=ENVS.CACHE_TTL)
    async def get_price(
        self, 
        token: Address, 
        block: Optional[Block] = None,
        ignore_pools: Tuple[Pool, ...] = (),
        ) -> Optional[UsdPrice]:

        quoter = await self.__quoter__(sync=False)
        if block and block < await contract_creation_block_async(quoter, True):
            return None

        paths = [[token, fee, usdc.address] for fee in self.fee_tiers]
        if token != weth:
            paths += [
                [token, fee, weth.address, self.fee_tiers[0], usdc.address] for fee in self.fee_tiers
            ]
        logger.debug("paths: %s", paths)
        
        amount_in = await ERC20(token, asynchronous=True).scale

        # TODO make this properly async after extending for brownie ContractTx
        results = await fetch_multicall(
            *[[quoter, 'quoteExactInput', self._encode_path(path), amount_in] for path in paths],
            block=block,
            sync=False
        )
        logger.debug("results: %s", results)
        # Quoter v2 uses this weird return struct, we must unpack it to get amount out.
        outputs = [
            (amount if isinstance(amount, int) else amount[0]) / self._undo_fees(path) / 1e6
            for amount, path in zip(results, paths)
            if amount
        ]
        logger.debug("outputs: %s", outputs)
        return UsdPrice(max(outputs)) if outputs else None

    @a_sync.aka.cached_property
    @stuck_coro_debugger
    async def pools(self) -> List[UniswapV3Pool]:
        factory = await self.__factory__(sync=False)
        return UniV3Pools(factory, asynchronous=self.asynchronous)

    @stuck_coro_debugger
    @a_sync.a_sync(ram_cache_maxsize=100_000, ram_cache_ttl=60*60)
    async def check_liquidity(self, token: Address, block: Block, ignore_pools: Tuple[Pool, ...] = ()) -> int:
        logger.debug("checking %s liquidity for %s at %s", self, token, block)
        if chain.id == Network.Mainnet and token == "0x6DEA81C8171D0bA574754EF6F8b412F2Ed88c54D":
            # LQTY, TODO refactor this somehow
            return 0
        
        quoter = await self.__quoter__(sync=False)
        if block and block < await contract_creation_block_async(quoter):
            logger.debug("block %s is before %s deploy block", block, quoter)
            return 0
        
        token_in_tasks: Dict[UniswapV3Pool, asyncio.Task] = {}
        token_out_tasks: Dict[UniswapV3Pool, asyncio.Task] = {}
        async for pool in self._pools_for_token(token, block):
            if pool not in ignore_pools:
                token_in_tasks[pool] = asyncio.create_task(pool.check_liquidity(token, block, sync=False))
                token_out_tasks[pool] = asyncio.create_task(pool.check_liquidity(pool._get_token_out(token), block, sync=False))
        
        # Since uni v3 liquidity can be provided asymmetrically, the most liquid pool in terms of `token` might not actually be the most liquid pool in terms of `token_out`
        # We need some spaghetticode here to account for these erroneous liquidity values
        # TODO: Refactor this
        token_out_liquidity: DefaultDict[ERC20, List[asyncio.Task]] = defaultdict(list)
        for pool, task in token_out_tasks.items():
            token_out_liquidity[pool._get_token_out(token)].append(task)
        
        token_out_min_liquidity = {token_out: min(await asyncio.gather(*tasks)) for token_out, tasks in token_out_liquidity.items()}
        for pool, task in token_out_tasks.items():
            token_out = pool._get_token_out(token)
            liquidity = await task
            if len(token_out_liquidity[token_out]) > 1 and liquidity == token_out_min_liquidity[token_out]:
                logger.debug("ignoring liquidity for %s", pool)
                token_in_tasks.pop(pool)
            elif token_out == weth and await task < 10 ** 19:  # 10 ETH
                # NOTE: this is totally arbitrary, works for all known cases but eventually will probably cause issues
                logger.debug("insufficient liquidity for %s", pool)
                token_in_tasks.pop(pool)

        liquidity = max(await asyncio.gather(*token_in_tasks.values())) if token_in_tasks else 0
        logger.debug("%s liquidity for %s at %s is %s", self, token, block, liquidity)
        return liquidity

    async def _pools_for_token(self, token: Address, block: Block) -> AsyncIterator[UniswapV3Pool]:
        pools = await self.__pools__(sync=False)
        async for pool in pools.objects(to_block=block):
            if token in pool:
                yield pool

    def _encode_path(self, path) -> bytes:
        types = [type for _, type in zip(path, cycle(['address', 'uint24']))]
        return encode_abi_packed(types, path)

    def _undo_fees(self, path) -> float:
        fees = [1 - fee / FEE_DENOMINATOR for fee in path if isinstance(fee, int)]
        return math.prod(fees)


class UniV3Pools(ProcessedEvents[UniswapV3Pool]):
    __slots__ = "asynchronous", 
    def __init__(self, factory: Contract, asynchronous: bool = False):
        self.asynchronous = asynchronous
        super().__init__(addresses=[factory.address], topics=[factory.topics["PoolCreated"]])
    def __repr__(self) -> str:
        return object.__repr__(self)
    def _process_event(self, event: _EventItem) -> UniswapV3Pool:
        token0, token1, fee, tick_spacing, pool = event.values()
        return UniswapV3Pool(pool, token0, token1, fee, tick_spacing, event.block_number, asynchronous=self.asynchronous)
    def _get_block_for_obj(self, obj: UniswapV3Pool) -> int:
        return obj._deploy_block


try:
    uniswap_v3 = UniswapV3(asynchronous=True)
except UnsupportedNetwork:
    uniswap_v3 = None



from tests.fixtures import mainnet_only
from y.prices import magic

@mainnet_only
def test_curve():
    token = "0xAA5A67c256e27A5d80712c51971408db3370927D" # DOLA3POOL3CRV-f
    assert magic.get_price(token, 15_000_000) == 1.0029303734731028

    token = "0x29059568bB40344487d62f7450E78b8E6C74e0e5" # YFIETH-f
    assert magic.get_price(token, 15_000_000) == 4713.219279160143




import asyncio
import logging
from decimal import Decimal
from typing import Optional

import a_sync

from y import convert
from y.classes.common import ERC20
from y.contracts import Contract, has_methods
from y.datatypes import AnyAddressType, Block, UsdPrice
from y.prices import magic

logger = logging.getLogger(__name__)

@a_sync.a_sync(default='sync', cache_type='memory')
async def is_ib_token(token: AnyAddressType) -> bool:
    return await has_methods(token, ('debtShareToVal(uint)(uint)','debtValToShare(uint)(uint)','token()(address)','totalToken()(uint)','totalSupply()(uint)'), sync=False)

@a_sync.a_sync(default='sync')
async def get_price(token: AnyAddressType, block: Optional[Block] = None) -> UsdPrice:
    address = convert.to_address(token)
    contract = await Contract.coroutine(address)
    token, total_bal, total_supply = await asyncio.gather(
        contract.token.coroutine(block_identifier=block),
        contract.totalToken.coroutine(block_identifier=block),
        contract.totalSupply.coroutine(block_identifier=block),
    )
    token_scale, pool_scale = await asyncio.gather(ERC20(token, asynchronous=True).scale, ERC20(address, asynchronous=True).scale)
    total_bal /= Decimal(token_scale)
    total_supply /= Decimal(pool_scale)
    share_price = total_bal / total_supply
    token_price = await magic.get_price(token, block, sync=False)
    price = share_price * token_price
    return price




import asyncio
import logging
from typing import Optional

from a_sync import a_sync
from brownie import convert
from pony.orm import commit, db_session

from y import constants
from y._db.common import token_attr_threads
from y._db.entities import Address, Token, insert, retry_locked
from y._db.utils._ep import _get_get_token
from y._db.utils.utils import get_chain
from y.erc20 import decimals

logger = logging.getLogger(__name__)

@a_sync(default='async', executor=token_attr_threads)
@db_session
@retry_locked
def get_token(address: str) -> Token:
    address = convert.to_address(address)
    if address == constants.EEE_ADDRESS:
        raise ValueError(f"cannot create token entity for {constants.EEE_ADDRESS}")
    while True:
        if entity := Address.get(chain=get_chain(sync=True), address=address):
            if isinstance(entity, Token):
                return entity
            entity.delete()
            commit()
        return insert(type=Token, chain=get_chain(sync=True), address=address) or Token.get(chain=get_chain(sync=True), address=address)
        
@a_sync(default='async', executor=token_attr_threads)
@db_session
@retry_locked
def get_bucket(address: str) -> Optional[str]:
    if address == constants.EEE_ADDRESS:
        return
    get_token = _get_get_token()
    return get_token(address, sync=True).bucket

@a_sync(default='async', executor=token_attr_threads)
@db_session
@retry_locked
def set_bucket(address: str, bucket: str) -> None:
    if address == constants.EEE_ADDRESS:
        return
    get_token = _get_get_token()
    get_token(address, sync=True).bucket = bucket

@a_sync(default='async', executor=token_attr_threads)
@db_session
@retry_locked
def get_symbol(address: str) -> Optional[str]:
    get_token = _get_get_token()
    return get_token(address, sync=True).symbol

@a_sync(default='async', executor=token_attr_threads)
@db_session
@retry_locked
def set_symbol(address: str, symbol: str) -> None:
    get_token = _get_get_token()
    get_token(address, sync=True).symbol = symbol

@a_sync(default='async', executor=token_attr_threads)
@db_session
@retry_locked
def get_name(address: str) -> Optional[str]:
    get_token = _get_get_token()
    return get_token(address, sync=True).name

@a_sync(default='async', executor=token_attr_threads)
@db_session
@retry_locked
def set_name(address: str, name: str) -> None:
    get_token = _get_get_token()
    get_token(address, sync=True).name = name

async def get_decimals(address: str) -> int:
    d = await _get_token_decimals(address)
    if d is None:
        d = await decimals(address, sync=False)
        if d:
            asyncio.create_task(coro=set_decimals(address, d), name=f"set_decimals {address}")
    return d

@a_sync(default='async', executor=token_attr_threads)
@db_session
@retry_locked
def _get_token_decimals(address: str) -> Optional[int]:
    get_token = _get_get_token()
    return get_token(address, sync=True).decimals

@a_sync(default='async', executor=token_attr_threads)
@db_session
@retry_locked
def set_decimals(address: str, decimals: int) -> None:
    get_token = _get_get_token()
    get_token(address, sync=True).decimals = decimals



import asyncio
import logging
from typing import List, Optional, Union

import a_sync
from brownie import chain

from y.datatypes import AnyAddressType, Block, UsdPrice
from y.decorators import stuck_coro_debugger
from y.networks import Network
from y.prices.dex.balancer.v1 import BalancerV1
from y.prices.dex.balancer.v2 import BalancerV2

logger = logging.getLogger(__name__)


class BalancerMultiplexer(a_sync.ASyncGenericBase):
    def __init__(self, asynchronous: bool = False) -> None:
        self.asynchronous = asynchronous
    
    def __repr__(self) -> str:
        return "<BalancerMuliplexer>"
    
    @a_sync.aka.property
    async def versions(self) -> List[Union[BalancerV1, BalancerV2]]:
        return [v for v in await asyncio.gather(self.v1, self.v2) if v]

    @a_sync.aka.cached_property
    async def v1(self) -> Optional[BalancerV1]:
        try: return BalancerV1(asynchronous=self.asynchronous)
        except ImportError: return None
    
    @a_sync.aka.cached_property
    async def v2(self) -> Optional[BalancerV2]:
        try: return BalancerV2(asynchronous=self.asynchronous)
        except ImportError: return None

    @stuck_coro_debugger
    async def is_balancer_pool(self, token_address: AnyAddressType) -> bool:
        return any(await asyncio.gather(*[v.is_pool(token_address, sync=False) for v in await self.versions]))
    
    @stuck_coro_debugger
    async def get_pool_price(self, token_address: AnyAddressType, block: Optional[Block] = None) -> Optional[UsdPrice]:
        versions: List[Union[BalancerV1, BalancerV2]] = await self.versions
        for v in versions:
            if await v.is_pool(token_address, sync=False):
                return UsdPrice(await v.get_pool_price(token_address, block, sync=False))
            
    @stuck_coro_debugger
    @a_sync.a_sync(cache_type='memory')
    async def get_price(self, token_address: AnyAddressType, block: Optional[Block] = None) -> Optional[UsdPrice]:
        if await self.is_balancer_pool(token_address, sync=False):
            return await self.get_pool_price(token_address, block=block, sync=False)

        price = None
        
        if ( # NOTE: Only query v2 if block queried > v2 deploy block plus some extra blocks to build up liquidity
            (chain.id == Network.Mainnet and (not block or block > 12272146 + 100000))
            or (chain.id == Network.Fantom and (not block or block > 16896080))
            ): 
            v2: BalancerV2 = await self.v2
            price = await v2.get_token_price(token_address, block, sync=False)
            if price:
                logger.debug("balancer v2 -> $%s", price)
                return price

        if not price and chain.id == Network.Mainnet:   
            v1: BalancerV1 = await self.v1   
            price = await v1.get_token_price(token_address, block, sync=False)
            if price:
                logger.debug("balancer v1 -> $%s", price)
                return price

balancer_multiplexer = BalancerMultiplexer(asynchronous=True)




from typing import Optional

import a_sync
from async_lru import alru_cache
from brownie import chain
from brownie.convert.datatypes import EthAddress

from y.classes.common import ERC20
from y.contracts import Contract
from y.datatypes import AnyAddressType
from y.networks import Network
from y.prices import magic
from y.exceptions import ContractNotVerified
from brownie.exceptions import ContractNotFound


@a_sync.a_sync(default='sync')
async def is_solidex_deposit(token: AnyAddressType) -> bool:
    if chain.id != Network.Fantom:
        return False
    try:
        contract = await Contract.coroutine(token)
    except (ContractNotVerified, ContractNotFound):
        return False
    if not hasattr(contract, 'pool'):
        return False
    name = await ERC20(token, asynchronous=True).name
    return name.startswith("Solidex") and name.endswith("Deposit")

@a_sync.a_sync(default='sync')
async def get_price(token: AnyAddressType, block: Optional[int] = None):
    pool = await _get_pool(token)
    return await magic.get_price(pool, block, sync=False)

@alru_cache(maxsize=None)
async def _get_pool(token) -> EthAddress:
    contract = await Contract.coroutine(token)
    return await contract.pool.coroutine()